#!/usr/bin/env python
#
# gllinevector.py - The GLLineVector class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLLineVector` class, for displaying 3D
vector :class:`.Image` overlays in line mode.

The :class:`.GLLineVertices` class is also defined in this module, and is used
when running in OpenGL 1.4. See the :mod:`.gl14.gllinevector_funcs` and
:mod:`.gl21.gllinevector_funcs` modules for more details.
"""

import logging

import numpy                as np

import fsl.data.dtifit      as dtifit
import fsleyes.gl           as fslgl
import fsleyes.gl.glvector  as glvector


log = logging.getLogger(__name__)


class GLLineVector(glvector.GLVector):
    """The ``GLLineVector`` class encapsulates the logic required to render an
    :class:`.Image` instance of shape ``x*y*z*3``, or type
    ``NIFTI_TYPE_RGB24``, as a vector image, where the vector at each voxel is
    drawn as a line, and coloured in the same way that voxels in the
    :class:`.GLRGBVector` are coloured.  The ``GLLineVector`` class assumes
    that the :class:`.Display` instance associated with the ``Image`` overlay
    holds a reference to a :class:`.LineVectorOpts` instance, which contains
    ``GLLineVector``-specific display settings.  The ``GLLineVector`` class is
    a sub-class of the :class:`.GLVector` class, and uses the functionality
    provided by ``GLVector``.


    In a similar manner to the :class:`.GLRGBVector`, the ``GLLineVector`` uses
    two OpenGL version-specific modules, the :mod:`.gl14.gllinevector_funcs`
    and :mod:`.gl21.gllinevector_funcs` modules. It is assumed that these
    modules define the same functions that are defined by the
    :class:`.GLRGBVector` version specific modules.


    A ``GLLineVector`` instance is rendered in different ways depending upon
    the rendering environment (GL 1.4 vs GL 2.1), so most of the rendering
    functionality is implemented in the version-specific modules mentioned
    above.
    """


    def __init__(self, image, overlayList, displayCtx, canvas, threedee):
        """Create a ``GLLineVector`` instance.

        :arg image:       An :class:`.Image` or :class:`.DTIFitTensor`
                          instance.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg canvas:      The canvas doing the drawing.
        :arg threedee:    2D or 3D rendering.
        """

        # If the overlay is a DTIFitTensor, use the
        # V1 image is the vector data. Otherwise,
        # assume that the overlay is the vector image.
        if isinstance(image, dtifit.DTIFitTensor): vecImage = image.V1()
        else:                                      vecImage = image

        def prefilter(data):
            # Scale to unit length if required.
            if not self.opts.unitLength:
                return data

            data = np.copy(data)

            with np.errstate(invalid='ignore'):

                # Vector images stored as RGB24 data
                # type are assumed to map from [0, 255]
                # to [-1, 1], so cannot be normalised
                if vecImage.nvals > 1:
                    return data

                # calculate lengths
                x    = data[0, ...]
                y    = data[1, ...]
                z    = data[2, ...]
                lens = np.sqrt(x ** 2 + y ** 2 + z ** 2)

                # scale lengths to 1
                data[0, ...] = x / lens
                data[1, ...] = y / lens
                data[2, ...] = z / lens

            return data

        def prefilterRange(dmin, dmax):
            if self.opts.unitLength:
                return 0, 1
            else:
                return dmin, dmax

        glvector.GLVector.__init__(self,
                                   image,
                                   overlayList,
                                   displayCtx,
                                   canvas,
                                   threedee,
                                   prefilter=prefilter,
                                   prefilterRange=prefilterRange,
                                   vectorImage=vecImage,
                                   init=lambda: fslgl.gllinevector_funcs.init(
                                       self))

        self.opts.addListener('lineWidth',  self.name, self.notify)
        self.opts.addListener('unitLength', self.name,
                              self.__unitLengthChanged)


    def destroy(self):
        """Must be called when this ``GLLineVector`` is no longer needed.
        Removes some property listeners from the :class:`.LineVectorOpts`
        instance, calls the OpenGL version-specific ``destroy``
        function, and calls the :meth:`.GLVector.destroy` method.
        """
        self.opts.removeListener('lineWidth',  self.name)
        self.opts.removeListener('unitLength', self.name)
        fslgl.gllinevector_funcs.destroy(self)
        glvector.GLVector.destroy(self)


    def getDataResolution(self, xax, yax):
        """Overrides :meth:`.GLImageObject.getDataResolution`. Returns a pixel
        resolution suitable for rendering this ``GLLineVector``.
        """

        res       = list(glvector.GLVector.getDataResolution(self, xax, yax))
        res[xax] *= 20
        res[yax] *= 20

        return res


    def compileShaders(self):
        """Overrides :meth:`.GLVector.compileShaders`. Calls the OpenGL
        version-specific ``compileShaders`` function.
        """
        fslgl.gllinevector_funcs.compileShaders(self)


    def updateShaderState(self):
        """Overrides :meth:`.GLVector.updateShaderState`. Calls the OpenGL
        version-specific ``updateShaderState`` function.
        """
        return fslgl.gllinevector_funcs.updateShaderState(self)


    def preDraw(self, xform=None, bbox=None):
        """Overrides :meth:`.GLVector.preDraw`. Calls the base class
        implementation, and then calls the OpenGL version-specific ``preDraw``
        function.
        """
        glvector.GLVector.preDraw(self, xform, bbox)
        fslgl.gllinevector_funcs.preDraw(self, xform, bbox)


    def draw2D(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.draw2D`. Calls the OpenGL
        version-specific ``draw2D`` function.
        """
        fslgl.gllinevector_funcs.draw2D(self, *args, **kwargs)


    def draw3D(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.draw3D`. Calls the OpenGL
        version-specific ``draw3D`` function.
        """
        fslgl.gllinevector_funcs.draw3D(self, *args, **kwargs)


    def drawAll(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.drawAll`. Calls the OpenGL
        version-specific ``drawAll`` function.
        """
        fslgl.gllinevector_funcs.drawAll(self, *args, **kwargs)


    def postDraw(self, xform=None, bbox=None):
        """Overrides :meth:`.GLVector.postDraw`. Calls the base class
        implementation, and then calls the OpenGL version-specific ``postDraw``
        function.
        """
        glvector.GLVector.postDraw(self, xform, bbox)
        fslgl.gllinevector_funcs.postDraw(self, xform, bbox)


    def __unitLengthChanged(self, *a):
        """Called when the :attr:`.LineVectorOptsunitLength` property
        changes. Refreshes the vector image texture data.
        """
        self.imageTexture.refresh()
        self.updateShaderState()


class GLLineVertices:
    """The ``GLLineVertices`` class is used when rendering a
    :class:`GLLineVector` with OpenGL 1.4. It contains logic to generate
    vertices for every vector in the vector :class:`.Image` that is being
    displayed by a ``GLLineVector`` instance.


    This class is used by the OpenGL 1.4 implementation - when using OpenGL
    2.1, the logic encoded in this class is implemented in the line vector
    vertex shader. This is because OpenGL 1.4 vertex programs (using the
    ARB_vertex_program extension) are unable to perform texture lookups,
    so cannot retrieve the vector data.


    After a ``GLLineVertices`` instance has been created, the :meth:`refresh`
    method can be used to generate line vector vertices and voxel
    coordinates for every voxel in the :class:`Image`. These vertices and
    coordinates are stored as attributes of the ``GLLineVertices`` instance.


    Later, when the line vectors from a 2D slice  of the image need to be
    displayed, the :meth:`getVertices2D` method can be used to extract the
    vertices and coordinates from the slice.


    A ``GLLineVertices`` instance is not associated with a specific
    ``GLLineVector`` instance. This is so that a single  ``GLLineVertices``
    instance can be shared between more than one ``GLLineVector``, avoiding
    the need to store multiple copies of the vertices and voxel
    coordinates. This means that a ``GLLineVector`` instance needs to be
    passed to most of the methods of a ``GLLineVertices`` instance.
    """

    def __init__(self, glvec):
        """Create a ``GLLineVertices``. Vertices are calculated for the
        given :class:`.GLLineVector` instance.

        :arg glvec: A :class:`GLLineVector` which is using this
                    ``GLLineVertices`` instance.
        """

        self.__hash = None
        self.refresh(glvec)


    def destroy(self):
        """Should be called when this ``GLLineVertices`` instance is no
        longer needed. Clears references to cached vertices/coordinates.
        """
        self.vertices  = None
        self.voxCoords = None


    def __hash__(self):
        """Returns a hash of this ``GLLineVertices`` instance. The hash value
        is calculated and cached on every call to :meth:`refresh`, using the
        :meth:`calculateHash` method.  This method returns that cached value.
        """
        return self.__hash


    def calculateHash(self, glvec):
        """Calculates and returns a hash value that can be used to determine
        whether the vertices of this this ``GLLineVertices`` instance need to
        be recalculated. The hash value is based on some properties of the
        :class:`.LineVectorOpts` instance, associated with the given
        :class:`.GLLineVector`.

        For a ``GLLineVertices`` instance called ``verts``, if the following
        test::

            hash(verts) != verts.calculateHash(glvec)

        evaluates to ``False``, the vertices need to be refreshed (via a
        call to :meth:`refresh`).
        """
        opts = glvec.opts
        return (hash(opts.transform)  ^
                hash(opts.orientFlip) ^
                hash(opts.directed)   ^
                hash(opts.unitLength) ^
                hash(opts.lengthScale))


    def refresh(self, glvec):
        """(Re-)calculates the vertices  of this ``GLLineVertices`` instance.

        For each voxel, in the :class:`.Image` overlay being displayed by the
        :class:`GLLineVector` associated with this ``GLLineVertices``
        instance, two vertices are generated, which define a line that
        represents the vector at the voxel.

        The vertices are stored as a :math:`X\\times Y\\times Z\\times
        2\\times 3` ``numpy`` array, as an attribute of this instance,
        called ``vertices``.
        """

        opts  = glvec.opts
        image = glvec.vectorImage
        data  = image.data
        shape = image.shape

        # Pull out the xyz components of the
        # vectors, and calculate vector lengths.

        # The image may either
        # have shape (X, Y, Z, 3)
        if image.nvals == 1:
            vertices = np.array(data, dtype=np.float32)
        # Or (we assume) a RGB
        # structured array
        else:
            vertices         = np.zeros(list(data.shape) + [3],
                                        dtype=np.float32)
            vertices[..., 0] = (data['R'].astype(np.float32) / 127.5) - 1
            vertices[..., 1] = (data['G'].astype(np.float32) / 127.5) - 1
            vertices[..., 2] = (data['B'].astype(np.float32) / 127.5) - 1

        x    = vertices[..., 0]
        y    = vertices[..., 1]
        z    = vertices[..., 2]
        lens = np.sqrt(x ** 2 + y ** 2 + z ** 2)

        # Flip vectors about the x axis if necessary
        if opts.orientFlip:
            x = -x

        if opts.unitLength:

            # scale the vector lengths to 0.5
            with np.errstate(invalid='ignore'):
                vertices[..., 0] = 0.5 * x / lens
                vertices[..., 1] = 0.5 * y / lens
                vertices[..., 2] = 0.5 * z / lens

            # Scale the vector data by the minimum
            # voxel length, so it is a unit vector
            # within real world space
            vertices /= (image.pixdim[:3] / min(image.pixdim[:3]))

        # Scale the vectors by the length scaling factor
        vertices *= opts.lengthScale / 100.0

        # Duplicate vector data so that each
        # vector is represented by two vertices,
        # representing a line through the origin.
        # Or, if displaying directed vectors,
        # add an origin point for each vector.
        if opts.directed:
            origins  = np.zeros(vertices.shape, dtype=np.float32)
            vertices = np.concatenate((origins, vertices), axis=3)
        else:
            vertices = np.concatenate((-vertices, vertices), axis=3)

        vertices = vertices.reshape((shape[0],
                                     shape[1],
                                     shape[2],
                                     2,
                                     3))

        self.vertices = vertices
        self.__hash   = self.calculateHash(glvec)


    def getVertices2D(self, glvec, zpos, axes, bbox=None):
        """Extracts and returns a slice of line vertices, and the associated
        voxel coordinates, which are in a plane located at the given Z
        position (in display coordinates).

        This method assumes that the :meth:`refresh` method has already been
        called.
        """

        image    = glvec.image
        shape    = image.shape[:3]
        xax, yax = axes[:2]

        vertices  = self.vertices
        voxCoords = glvec.generateVoxelCoordinates2D(zpos, axes, bbox)

        # Turn the voxel coordinates into
        # indices suitable for looking up
        # the corresponding vertices
        coords = np.array(np.floor(voxCoords + 0.5), dtype=np.int32)

        # remove any out-of-bounds voxel coordinates
        shape     = np.array(vertices.shape[:3])
        inBounds  = ((coords >= [0, 0, 0]) & (coords < shape)).all(1)
        coords    = coords[   inBounds, :].T
        voxCoords = voxCoords[inBounds, :]

        # pull out the vertex data
        # corresponding to the voxels
        vertices = vertices[coords[0], coords[1], coords[2], :, :]
        vertices = vertices.reshape(-1, 3)

        if not vertices.flags['C_CONTIGUOUS']:
            vertices = np.ascontiguousarray(vertices)

        voxCoords = voxCoords.repeat(repeats=2, axis=0)

        return vertices, voxCoords



#!/usr/bin/env python
#
# state.py - Functions for saving/restoring the state of *FSLeyes*.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides two functions, :func:`getState` and :func:`setState`.
These functions may be used to get/set the state of *FSLeyes*.
"""


import logging

import fsleyes.actions.showcommandline  as showcommandline
import fsleyes.actions.applycommandline as applycommandline
import fsleyes.views.canvaspanel        as canvaspanel
import fsleyes.layouts                  as layouts


log = logging.getLogger(__name__)


SEP = '\n------\n'
"""String used to separate different parts of the state string. """


CLISEP = '\t'
"""String used to separate different command line arguments, within the state
string.
"""


def getState(frame):
    """Generate a string which describes the current state of FSLeyes.

    The string contains:

      - A layout specification string, as generated by the
        :func:`.layouts.serialiseLayout` function.

      - A set of command line arguments describing the overlays that are
        loaded, generated by the :func:`.genCommandLineArgs` function.

      - A set of command line arguments for each :class:`.CanvasPanel`,
        describing the panel settings, and any panel-specific overlay
        settings.

    :arg frame: The :class:`.FSLeyesFrame`
    :returns:   A string describing FSLeyes.
    """

    displayCtx  = frame.displayCtx
    overlayList = frame.overlayList

    layout = layouts.serialiseLayout(frame)
    cli    = [showcommandline.genCommandLineArgs(overlayList, displayCtx)]

    for i, panel in enumerate(frame.viewPanels):
        if not isinstance(panel, canvaspanel.CanvasPanel):
            continue

        argv = showcommandline.genCommandLineArgs(
            overlayList,
            panel.displayCtx,
            panel)
        cli.append(argv)

    cli   = [[a.strip('"') for a in argv[1:]] for argv in cli]
    cli   = [CLISEP.join(c) for c in cli]
    state = SEP.join([layout] + cli)

    return state


def setState(frame, state):
    """Set the state of FSLeyes from the given ``state`` string.

    .. warning:: This function will remove all view panels, and remove all
                 overlays, before loading the new state.

    :arg frame: The :class:`.FSLeyesFrame`
    :arg state: A FSLeyes state string, generated by :func:`getState`.
    """

    displayCtx  = frame.displayCtx
    overlayList = frame.overlayList

    bits   = state.split(SEP)
    layout = bits[0]
    cli    = bits[1:]
    cli    = [c.split(CLISEP) for c in cli]

    # First clear the current state
    frame.removeAllViewPanels()
    overlayList.clear()

    # Then load all new overlays
    applycommandline.applyCommandLineArgs(
        overlayList,
        displayCtx,
        cli[0],
        blocking=True)

    # After overlays have been loaded,
    # apply the new layout, and apply
    # view-specific overlay arguments.
    layouts.applyLayout(frame, 'state', layout)

    panels = [p for p in frame.viewPanels
              if isinstance(p, canvaspanel.CanvasPanel)]
    for i, panel in enumerate(panels, 1):
        applycommandline.applyCommandLineArgs(
            overlayList,
            panel.displayCtx,
            cli[i],
            panel,
            loadOverlays=False)



#!/usr/bin/env python
#
# overlay.py - Defines the OverlayList class, and a few utility functions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module defines the :class:`OverlayList` class, which is a simple but
fundamental class in *FSLeyes* - it is a container for all loaded overlays.
Only one ``OverlayList`` ever exists, and it is shared throughout the entire
application.


**What is an overlay?**


The definition of an *overlay* is fairly broad; any object can be added to
the ``OverlayList``  - there is no ``Overlay`` base class, nor any interface
which must be provided by an overlay object. The only requirements imposed on
an overlay type are:

  - Must be able to be created with a single ``__init__`` parameter, which
    is a string specifying the data source location (e.g. a file name) (but
    see the note below about ``Image`` overlays).

  - Must have an attribute called ``name``, which is used as the initial
    display name for the overlay.

  - Must have an attribute called ``dataSource``, which is used to identify
    the source of the overlay data.

  - Must be hashable (i.e. usable as a dictionary key).

  - Must be supported by the :mod:`~fsleyes.gl` package .. ok, this is a
    pretty big requirement .. See the :mod:`.globject` and the
    :data:`.displaycontext.OVERLAY_TYPES` documentation for details on how to
    get started with this one.


One further requirement is imposed on overlay types which derive from the
:class:`.Image` class:

 -  The ``__init__`` method for sub-classes of the ``Image`` class must
    accept the ``loadData``, ``calcRange``, ``indexed`` , and ``threaded``
    parameters, and pass them through to the base class ``__init__`` method.


Currently (``fsleyes`` version |version|) the only overlay types in existence
(and able to be rendered) are:

.. autosummary::
   :nosignatures:

   ~fsl.data.image.Image
   ~fsl.data.featimage.FEATImage
   ~fsl.data.melodicimage.MelodicImage
   ~fsl.data.mghimage.MGHImage
   ~fsl.data.dtifit.DTIFitTensor
   ~fsl.data.vtk.VTKMesh
   ~fsl.data.gifti.GiftiMesh
   ~fsl.data.freesurfer.FreesurferMesh
   ~fsl.data.bitmap.Bitmap


This module also provides a few convenience classes and functions:


.. autosummary::
   :nosignatures:

   ProxyImage
   findFEATImage
"""


import os.path as op
import            logging
import            weakref
import            collections

import fsl.data.utils as dutils
import fsl.data.image as fslimage
import fsleyes_props  as props


log = logging.getLogger(__name__)


class OverlayList(props.HasProperties):
    """Class representing a collection of overlays to be displayed together.

    Contains a :class:`props.properties_types.List` property called
    :attr:`overlays`, containing overlay objects (e.g. :class:`.Image` or
    :class:`.Mesh` objects). Listeners can be registered on the
    ``overlays`` property, so they are notified when the overlay list changes.

    An :class:`OverlayList` object has a few wrapper methods around the
    :attr:`overlays` property, allowing the :class:`OverlayList` to be used as
    if it were a list itself.

    The :mod:`.loadoverlay` module contains some convenience functions for
    loading and adding overlays.

    The :meth:`getData` and :meth:`setData` methods allow arbitrary bits
    of data associated with an overlay to be stored and retrieved.
    """


    def __validateOverlay(self, atts, overlay):
        """Makes sure that the given overlay object is valid."""
        return (hasattr(overlay, 'name')      and
                hasattr(overlay, 'dataSource'))


    overlays = props.List(
        listType=props.Object(allowInvalid=False,
                              validateFunc=__validateOverlay))
    """A list of overlay objects to be displayed."""


    def __init__(self, overlays=None):
        """Create an ``OverlayList`` object from the given sequence of
        overlays."""

        if overlays is None: overlays = []
        self.overlays.extend(overlays)

        # The append/insert methods allow display/opts
        # property values to be specified for newly
        # added overlays. These can be queried via
        # the initProps method (and is done so by
        # DisplayContext/Display instances).
        # It is a dict of dicts:
        #
        #    {
        #      overlay : {
        #        propName: value,
        #        propName: value,
        #      },
        #      ...
        #    }
        self.__initProps = collections.defaultdict(dict)

        # This dictionary may be used throughout FSLeyes,
        # via the getData/setData methods, to store
        # any sort of data associated with an overlay.
        # It is a dict of dicts:
        #
        #   {
        #      overlay : {
        #        key : value,
        #        key : value,
        #      },
        #      overlay : {
        #        key : value,
        #        key : value,
        #      }
        #   }
        self.__overlayData = weakref.WeakKeyDictionary()


    def initProps(self, overlay):
        """Returns a dict containing initial :class:`.Display` and
        :class:`.DisplayOpts` property values to be used for the given
        ``overlay``, if they were specified via the :meth:`append` or
        :meth:`insert` methods.

        This method requires that there is no overlap between the property
        names used in :class:`.Display` and :class:`.DisplayOpts` classes.
        """
        return self.__initProps[overlay]


    def getData(self, overlay, key, *args):
        """Returns any stored value associated with the specified ``overlay``
        and ``key``.

        :arg default: Default value if there is no value associated with the
                      given ``key``. If not specified, and an unknown key is
                      given, a ``KeyError`` is raised.
        """
        if len(args) not in (0, 1):
            raise RuntimeError('Invalid arguments: {}'.format(args))

        ovlDict = self.__overlayData.get(overlay, {})

        if len(args) == 1:
            return ovlDict.get(key, args[0])
        else:
            return ovlDict[key]


    def setData(self, overlay, key, value):
        """Stores the given value via the specified ``overlay`` and ``key``.
        """
        ovlDict = self.__overlayData.get(overlay, None)

        if ovlDict is not None:
            ovlDict[key] = value
        else:
            self.__overlayData[overlay] = {key : value}


    def find(self, name):
        """Returns the first overlay with the given ``name`` or ``dataSource``,
        or ``None`` if there is no overlay with said ``name``/``dataSource``.
        """

        if name is None:
            return None

        absname = op.abspath(name)

        for overlay in self.overlays:

            if overlay.name == name:
                return overlay

            if overlay.dataSource is None:
                continue

            # Ignore file extensions for NIFTI images.
            if isinstance(overlay, fslimage.Image):
                if fslimage.removeExt(overlay.dataSource) == \
                   fslimage.removeExt(absname):
                    return overlay
            else:
                if overlay.dataSource == absname:
                    return overlay

        return None


    def __str__(self):
        return self.overlays.__str__()

    def __repr__(self):
        return self.overlays.__str__()


    # Wrappers around the overlays list property, allowing this
    # OverlayList object to be used as if it is actually a list.
    def __len__(self):
        return self.overlays.__len__()

    def __getitem__(self, key):
        return self.overlays.__getitem__(key)

    def __iter__(self):
        return self.overlays.__iter__()

    def __contains__(self, item):
        return self.overlays.__contains__(item)

    def __setitem__(self, key, val):
        return self.overlays.__setitem__(key, val)

    def __delitem__(self, key):

        if   isinstance(key, slice): pass
        elif isinstance(key, int):   key = slice(key, key + 1, None)
        else:                        raise IndexError('Invalid key type')

        ovls = self[key]
        for ovl in ovls:
            self.__initProps.pop(ovl, None)

        return self.overlays.__delitem__(key)

    def index(self, item):
        return self.overlays.index(item)

    def count(self, item):
        return self.overlays.count(item)

    def insert(self, index, item, **initProps):
        """Insert a new overlay into the overlay list.

        Any initial :class:`.Display`/:class:`.DisplayOpts` property values
        may be passed in as keyword arguments.
        """

        with props.suppress(self, 'overlays', notify=True):
            self.overlays.insert(index, item)
            self.__initProps[item] = initProps

    def append(self, item, **initProps):
        """Add a new overlay to the end of the overlay list.

        Any initial :class:`.Display`/:class:`.DisplayOpts` property values
        may be passed in as keyword arguments.
        """
        self.insert(len(self), item, **initProps)

    def extend(self, iterable, **initProps):
        """Add new overlays to the overlay list.

        Any initial :class:`.Display`/:class:`.DisplayOpts` property values
        may be passed in as keyword arguments, where the argument name is the
        property name, and the argument value is a dict of
        ``{overlay : value}`` mappings.
        """

        with props.suppress(self, 'overlays', notify=True):

            result = self.overlays.extend(iterable)

            for propName, overlayProps in initProps.items():
                for overlay, val in overlayProps.items():
                    self.__initProps[overlay][propName] = val

        return result

    def insertAll(self, index, items):
        return self.overlays.insertAll(index, items)

    def pop(self, index=-1):
        ovl = self.overlays.pop(index)
        self.__initProps.pop(ovl, None)
        return ovl

    def move(self, from_, to):
        return self.overlays.move(from_, to)

    def remove(self, item):
        self.overlays.remove(item)
        self.__initProps.pop(item, None)

    def clear(self):
        del self[:]


class ProxyImage(fslimage.Image):
    """The ``ProxyImage`` class is a simple wrapper around an :class:`Image`
    instance. It is intended to be used to represent images or data which
    are derived from another image.
    """

    def __init__(self, base, *args, **kwargs):
        """Create a ``ProxyImage``.

        :arg base:   The :class:`Image` instance upon which this ``ProxyImage``
                     is based.

        :arg volume: Must be passed as a keyword argument. If provided, is a
                     slice into the image data specifying the 3D volume to
                     copy. If not provided, the entire image is copied.
        """

        if not isinstance(base, fslimage.Image):
            raise ValueError('Base image must be an Image instance')

        self.__base = base

        kwargs['header'] = base.header

        volume = kwargs.pop('volume', None)
        if volume is not None: data = base[volume]
        else:                  data = base[:]

        fslimage.Image.__init__(self, data, *args, **kwargs)


    def getBase(self):
        """Returns the base :class:`Image` of this ``ProxyImage``. """
        return self.__base


def findFEATImage(overlayList, overlay):
    """Searches the given :class:`.OverlayList` to see if there is a
    :class:`.FEATImage` associated with the given ``overlay``. Returns the
    ``FEATImage`` if found, otherwise returns ``None``.
    """

    import fsl.data.featanalysis as featanalysis
    import fsl.data.featimage    as featimage

    if isinstance(overlay, featimage.FEATImage): return overlay
    if overlay            is None:               return None
    if overlay.dataSource is None:               return None

    featPath = featanalysis.getAnalysisDir(overlay.dataSource)

    if featPath is None:
        return None

    dataPath  = featanalysis.getDataFile(featPath)
    featImage = overlayList.find(dataPath)

    return featImage


def findMeshReferenceImage(overlayList, overlay):
    """Searches the :class:`.OverlayList` and tries to identify a reference
    image for the given :class:`.Mesh` overlay. Returns the identified
    overlay, or ``None`` if one can't be found.
    """

    import fsl.data.vtk as fslvtk

    # TODO support for gifti/freesurfer

    if not isinstance(overlay, fslvtk.VTKMesh):
        return None

    try:
        prefix = fslvtk.getFIRSTPrefix(overlay.dataSource)

        for ovl in overlayList:
            if prefix.startswith(ovl.name):
                return ovl

    except Exception:
        pass

    return None



#!/usr/bin/env python
#
# glvolume_funcs.py - OpenGL 1.4 functions used by the GLVolume class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLVolume`
class to render :class:`.Image` overlays in an OpenGL 1.4 compatible manner.

An :class:`.ARBPShader` is used to manage the ``glvolume`` vertex/fragment
programs.
"""


import logging

import numpy                as np
import OpenGL.GL            as gl

import fsl.transform.affine as affine
import fsleyes.gl.shaders   as shaders
import fsleyes.gl.routines  as glroutines
import fsleyes.gl.glvolume  as glvolume


log = logging.getLogger(__name__)


def init(self):
    """Calls :func:`compileShaders` and :func:`updateShaderState`."""

    self.shader = None

    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Deletes handles to the vertex/fragment programs."""

    if self.shader is not None:
        self.shader.destroy()
        self.shader = None


def compileShaders(self):
    """Creates a :class:`.ARBPShader` instance. """

    if self.shader is not None:
        self.shader.destroy()

    if self.threedee: frag = 'glvolume_3d'
    else:             frag = 'glvolume'

    vertSrc  = shaders.getVertexShader(  'glvolume')
    fragSrc  = shaders.getFragmentShader(frag)
    texes    = {
        'imageTexture'     : 0,
        'colourTexture'    : 1,
        'negColourTexture' : 2,
        'clipTexture'      : 3,
        'modulateTexture'  : 4
    }

    constants = {
        'kill_fragments_early' : not self.threedee,
        'texture_is_2d'        : self.imageTexture.ndim == 2
    }

    if self.threedee:

        if   self.opts.clipMode == 'intersection': clipMode = 1
        elif self.opts.clipMode == 'union':        clipMode = 2
        elif self.opts.clipMode == 'complement':   clipMode = 3
        else:                                      clipMode = 0

        constants['numSteps']        = self.opts.numInnerSteps
        constants['clipMode']        = clipMode
        constants['numClipPlanes']   = self.opts.numClipPlanes
        texes[    'startingTexture'] = 5
        texes[    'depthTexture']    = 6

    self.shader = shaders.ARBPShader(vertSrc,
                                     fragSrc,
                                     shaders.getShaderDir(),
                                     texes,
                                     constants)


def updateShaderState(self):
    """Sets all variables required by the vertex and fragment programs. """

    if not self.ready():
        return

    opts   = self.opts
    shader = self.shader

    # enable the vertex and fragment programs
    shader.load()

    # The voxValXform transformation turns
    # an image texture value into a raw
    # voxel value. The colourMapXform
    # transformation turns a raw voxel value
    # into a value between 0 and 1, suitable
    # for looking up an appropriate colour
    # in the 1D colour map texture.
    voxValXform = affine.concat(self.colourTexture.getCoordinateTransform(),
                                self.imageTexture.voxValXform)
    voxValXform = [voxValXform[0, 0], voxValXform[0, 3], 0, 0]

    # And the clipping range, normalised
    # to the image texture value range
    invClip     = 1 if opts.invertClipping    else -1
    useNegCmap  = 1 if opts.useNegativeCmap   else  0
    imageIsClip = 1 if opts.clipImage is None else -1

    # modalpha not applied in 3D
    modAlpha    = 1 if opts.modulateAlpha         else -1
    imageIsMod  = 1 if opts.modulateImage is None else -1
    modXform    = self.getModulateValueXform()

    imgXform = self.imageTexture.invVoxValXform
    if opts.clipImage is None: clipXform = imgXform
    else:                      clipXform = self.clipTexture.invVoxValXform

    clipLo  = opts.clippingRange[0] * clipXform[0, 0] + clipXform[0, 3]
    clipHi  = opts.clippingRange[1] * clipXform[0, 0] + clipXform[0, 3]
    texZero = 0.0                   * imgXform[ 0, 0] + imgXform[ 0, 3]

    clipping = [clipLo, clipHi, invClip, imageIsClip]
    modulate = [modXform[0, 0], modXform[0, 3], modAlpha, imageIsMod]
    negCmap  = [useNegCmap, texZero, modAlpha, 0]

    # disable clip image/modalpha for 3D
    if self.threedee:
        clipping[3] =  1
        modulate[2] = -1
        modulate[3] =  1

    changed  = False
    changed |= shader.setFragParam('voxValXform', voxValXform)
    changed |= shader.setFragParam('clipping',    clipping)
    changed |= shader.setFragParam('modulate',    modulate)
    changed |= shader.setFragParam('negCmap',     negCmap)

    if self.threedee:
        clipPlanes  = np.zeros((5, 4), dtype=np.float32)
        d2tmat      = opts.getTransform('display', 'texture')

        for i in range(opts.numClipPlanes):
            origin, normal   = self.get3DClipPlane(i)
            origin           = affine.transform(origin, d2tmat)
            normal           = affine.transformNormal(normal, d2tmat)
            clipPlanes[i, :] = glroutines.planeEquation2(origin, normal)

        changed |= shader.setFragParam('clipPlanes', clipPlanes)

    self.shader.unload()

    return changed


def preDraw(self, xform=None, bbox=None):
    """Prepares to draw a slice from the given :class:`.GLVolume` instance. """

    self.shader.load()
    self.shader.loadAtts()

    if isinstance(self, glvolume.GLVolume):
        clipCoordXform = self.getAuxTextureXform('clip')
        modCoordXform  = self.getAuxTextureXform('modulate')
        self.shader.setVertParam('clipCoordXform', clipCoordXform)
        self.shader.setVertParam('modCoordXform',  modCoordXform)


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Draws a 2D slice of the image at the given Z location. """

    vertices, voxCoords, texCoords = self.generateVertices2D(
        zpos, axes, bbox=bbox)

    if xform is not None:
        vertices = affine.transform(vertices, xform)

    vertices = np.array(vertices, dtype=np.float32).ravel('C')

    # Voxel coordinates are calculated
    # in the vertex program
    self.shader.setAtt('texCoord', texCoords)

    with glroutines.enabled((gl.GL_VERTEX_ARRAY)):
        gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)
        gl.glDrawArrays(gl.GL_TRIANGLES, 0, 6)


def draw3D(self, xform=None, bbox=None):
    """Draws the image in 3D on the canvas.

    :arg self:    The :class:`.GLVolume` object which is managing the image
                  to be drawn.

    :arg xform:   A 4*4 transformation matrix to be applied to the vertex
                  data.

    :arg bbox:    An optional bounding box.
    """
    opts    = self.opts
    canvas  = self.canvas
    display = self.display
    shader  = self.shader
    shape   = self.image.shape
    proj    = canvas.projectionMatrix
    src     = self.renderTexture1
    dest    = self.renderTexture2
    w, h    = src.shape

    vertices, voxCoords, texCoords = self.generateVertices3D(bbox)
    rayStep, texform               = opts.calculateRayCastSettings(xform, proj)

    rayStep = affine.transformNormal(
        rayStep, self.imageTexture.texCoordXform(shape))
    texform = affine.concat(
        texform, self.imageTexture.invTexCoordXform(shape))

    if xform is not None:
        vertices = affine.transform(vertices, xform)

    vertices = np.array(vertices, dtype=np.float32).ravel('C')

    outerLoop  = opts.getNumOuterSteps()
    screenSize = [
        1.0 / w,
        1.0 / h,
        1 if opts.blendByIntensity else -1,
        0]
    rayStep    = list(rayStep)   + [0]
    texform    = texform[2, :]
    settings   = [
        opts.blendFactor,
        0,
        0,
        display.alpha / 100.0]

    gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)

    shader.setAtt(      'texCoord',        texCoords)
    shader.setFragParam('rayStep',         rayStep)
    shader.setFragParam('screenSize',      screenSize)
    shader.setFragParam('tex2ScreenXform', texform)

    # Disable blending - we want each
    # loop to replace the contents of
    # the texture, not blend into it!
    with glroutines.enabled((gl.GL_VERTEX_ARRAY)), \
         glroutines.disabled((gl.GL_BLEND)):

        # The depth value for a fragment will
        # not necessary be set at the same
        # time that the fragment colour is
        # set, so we need to use <= for depth
        # testing so that unset depth values
        # do not cause depth clipping.
        gl.glDepthFunc(gl.GL_LEQUAL)

        for i in range(outerLoop):

            settings    = list(settings)
            dtex        = src.depthTexture
            settings[1] = i * opts.numInnerSteps

            if i == outerLoop - 1: settings[2] =  1
            else:                  settings[2] = -1

            shader.setFragParam('settings', settings)

            dest.bindAsRenderTarget()
            src .bindTexture(gl.GL_TEXTURE5)
            dtex.bindTexture(gl.GL_TEXTURE6)

            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            gl.glDrawArrays(gl.GL_TRIANGLES, 0, 36)

            src .unbindTexture()
            dtex.unbindTexture()
            dest.unbindAsRenderTarget()

            dest, src = src, dest

    gl.glDepthFunc(gl.GL_LESS)

    shader.unloadAtts()
    shader.unload()

    self.renderTexture1 = src
    self.renderTexture2 = dest


def drawAll(self, axes, zposes, xforms):
    """Draws mutltiple slices of the given image at the given Z position,
    applying the corresponding transformation to each of the slices.
    """

    nslices   = len(zposes)
    vertices  = np.zeros((nslices * 6, 3), dtype=np.float32)
    texCoords = np.zeros((nslices * 6, 3), dtype=np.float32)
    indices   = np.arange(nslices * 6,     dtype=np.uint32)

    for i, (zpos, xform) in enumerate(zip(zposes, xforms)):

        v, vc, tc = self.generateVertices2D(zpos, axes)

        vertices[ i * 6: i * 6 + 6, :] = affine.transform(v, xform)
        texCoords[i * 6: i * 6 + 6, :] = tc

    vertices = vertices.ravel('C')

    self.shader.setAtt('texCoord', texCoords)

    with glroutines.enabled((gl.GL_VERTEX_ARRAY)):
        gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)
        gl.glDrawElements(gl.GL_TRIANGLES,
                          nslices * 6,
                          gl.GL_UNSIGNED_INT,
                          indices)


def postDraw(self, xform=None, bbox=None):
    """Cleans up the GL state after drawing from the given :class:`.GLVolume`
    instance.
    """
    self.shader.unloadAtts()
    self.shader.unload()

    if self.threedee:
        self.drawClipPlanes(xform=xform)



#!/usr/bin/env python
#
# gllinevector_funcs.py - OpenGL 2.1 functions used by the GLLineVector class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLLineVector`
class to render :class:`.Image` overlays as line vector images in an OpenGL 2.1
compatible manner.


This module uses functions in the :mod:`.gl21.glvector_funcs` module, which
contains logic used for rendering both ``GLRGBVector`` and ``GLLineVector``
instances.


The voxel coordinates for every vector are passed directly to a vertex shader
program which calculates the position of the corresponding line vertices.


The ``glvector`` fragment shader (the same as that used by the
:class:`.GLRGBVector` class) is used to colour each line according to the
orientation of the underlying vector.
"""


from __future__ import division


import logging

import numpy               as np
import OpenGL.GL           as gl

import fsl.data.constants   as constants
import fsl.transform.affine as affine
from . import                  glvector_funcs


log = logging.getLogger(__name__)


def init(self):
    """Compiles and configures the vertex/fragment shaders used to render the
    ``GLLineVector`` via calls to :func:`compileShaders` and
    :func:`updateShaderState`.
    """

    self.shader = None

    name = self.name

    compileShaders(   self)
    updateShaderState(self)

    def update(*a, **kwa):
        updateShaderState(self)
        self.notify()

    # GLVector.addListener adds a listener
    # for the transform property, so we
    # overwrite it here - we need to update
    # the display<->voxel transformation
    # matrices whenever the transform
    # changes.
    self.opts.addListener('orientFlip',  name, update, weak=False)
    self.opts.addListener('directed',    name, update, weak=False)
    self.opts.addListener('lengthScale', name, update, weak=False)
    self.opts.addListener('transform',
                          name,
                          update,
                          overwrite=True,
                          weak=False)


def destroy(self):
    """Deletes the vertex/fragment shaders. """

    self.opts.removeListener('orientFlip',  self.name)
    self.opts.removeListener('directed',    self.name)
    self.opts.removeListener('lengthScale', self.name)
    self.opts.removeListener('transform',   self.name)

    self.shader.destroy()


def compileShaders(self):
    """Compiles the vertex/fragment shaders via the
    :func:`.gl21.glvector_funcs.compileShaders` function.
    """

    self.shader = glvector_funcs.compileShaders(self, 'gllinevector')


def updateShaderState(self):
    """Updates all variables used by the vertex/fragment shaders. The fragment
    shader is configured by the
    :func:`.gl21.glvector_funcs.updateFragmentShaderState` function.
    """

    shader = self.shader
    shader.load()

    changed     = glvector_funcs.updateShaderState(self)
    image       = self.vectorImage
    opts        = self.opts

    # see comments in gl21/glvector_funcs.py
    if self.vectorImage.niftiDataType == constants.NIFTI_DT_RGB24:
        vvxMat = affine.scaleOffsetXform(2, -1)
    else:
        vvxMat = self.imageTexture.voxValXform

    directed    = opts.directed
    lengthScale = opts.lengthScale / 100.0
    imageDims   = image.pixdim[:3]
    d2vMat      = opts.getTransform('display', 'voxel')
    v2dMat      = opts.getTransform('voxel',   'display')
    xFlip       = opts.orientFlip

    # If the unitLength option is on, the vector
    # data will have already been scaled to have
    # length 1 (see GLLineVector.__init__). But
    # we draw vectors in two parts, from the voxel
    # centre. So we have to half the vector lengths.
    if opts.unitLength:
        lengthScale /= 2
        # We also scale the vector data by the
        # minimum voxel length, so that each
        # vector has unit length relative to
        # the voxel dimensions.
        fac          = (image.pixdim[:3] / min(image.pixdim[:3]))
        lengthScale /= fac

    changed |= shader.set('vectorTexture',   4)
    changed |= shader.set('displayToVoxMat', d2vMat)
    changed |= shader.set('voxToDisplayMat', v2dMat)
    changed |= shader.set('voxValXform',     vvxMat)
    changed |= shader.set('imageDims',       imageDims)
    changed |= shader.set('directed',        directed)
    changed |= shader.set('lengthScale',     lengthScale)
    changed |= shader.set('xFlip',           xFlip)

    shader.unload()

    return changed


def preDraw(self, xform=None, bbox=None):
    """Prepares the GL state for drawing. This amounts to loading the
    vertex/fragment shader programs.
    """
    self.shader.load()


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Draws the line vectors at a plane at the specified Z location.
    Voxel coordinates are passed to the vertex shader, which calculates
    the corresponding line vertex locations.
    """

    opts   = self.opts
    shader = self.shader
    v2dMat = opts.getTransform('voxel', 'display')

    voxels  = self.generateVoxelCoordinates2D(zpos, axes, bbox=bbox)
    voxels  = np.repeat(voxels, 2, 0)
    indices = np.arange(voxels.shape[0], dtype=np.uint32)

    if xform is None: xform = v2dMat
    else:             xform = affine.concat(xform, v2dMat)

    shader.set(   'voxToDisplayMat', xform)
    shader.setAtt('vertexID',        indices)
    shader.setAtt('voxel',           voxels)
    shader.loadAtts()

    gl.glLineWidth(opts.lineWidth)
    gl.glDrawArrays(gl.GL_LINES, 0, voxels.size // 3)

    shader.unloadAtts()


def draw3D(self, xform=None, bbox=None):
    """Draws the line vectors in 3D space. """
    pass


def drawAll(self, axes, zposes, xforms):
    """Draws the line vectors at every slice specified by the Z locations. """

    for zpos, xform in zip(zposes, xforms):
        self.draw2D(zpos, axes, xform)


def postDraw(self, xform=None, bbox=None):
    """Clears the GL state after drawing. """
    self.shader.unload()



#!/usr/bin/env python
#
# loaddicom.py - The LoadDicomAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadDicomAction` class, an :class:`.Action`
which allows the user to load images from a DICOM directory.

A few standalone functions are also defined in this module:

.. autosummary::
   :nosignatures:

   loadDicom
"""


import                 os
import os.path      as op
from   datetime import datetime

import wx

import fsleyes_widgets.widgetgrid     as wg
import fsleyes_widgets.utils.status   as status
import fsleyes_widgets.utils.progress as progress
import fsl.utils.settings             as fslsettings
import fsl.data.dicom                 as fsldcm
import fsleyes.strings                as strings
import fsleyes.autodisplay            as autodisplay
from . import                            base


class LoadDicomAction(base.Action):
    """The ``LoadDicomAction`` is an :class:`.Action` which allows the user to
    load images from a DICOM directory. When invoked, the ``LoadDicomAction``
    does the following:

    1. Prompts the user to select a DICOM directory
    2. Identifies the data series that are present in the directory
    3. Prompts the user to select which series they would like to load
    4. Loads the selected series.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``LoadDicomAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__loadDicom)
        self.__frame = frame

        # permanently disable if
        # fsl.data.dicom says it
        # is disabled
        self.enabled = fsldcm.enabled()


    def __loadDicom(self):
        """Called when this ``LoadDicomAction`` is invoked. Calls the
        :func:`loadDicom` function.
        """

        def onLoad(overlays):
            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)

            self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[-1]

            if self.displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.overlayList,
                                            self.displayCtx)

        loadDicom(parent=self.__frame, callback=onLoad)


def loadDicom(dcmdir=None, parent=None, callback=None):
    """Does the following:

      1. Prompts the user to select a DICOM directory (unless
         ``dcmdir is not None``)

      2. Loads metadata about all of the data series in the
         DICOM directory

      3. Uses a :class:`.BrowseDicomDialog` to allow the user
         to choose which data series they wish to load

      4. Loads the selected series, and passes them to the
         ``callback`` function if it is provided.

    :arg dcmdir:   Directory to load DICOMs from. If not provided, the user is
                   prompted to select a directory.

    :arg parent:   ``wx`` parent object.

    :arg callback: Function which is passed the loaded DICOM series
                   (:class:`.Image` objects).
    """

    if parent is None:
        parent = wx.GetTopLevelWindows()[0]

    # 1. prompt user to select dicom directory
    if dcmdir is None:
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        dlg     = wx.DirDialog(
            parent,
            message=strings.messages['loadDicom.selectDir'],
            defaultPath=fromDir,
            style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)

        if dlg.ShowModal() != wx.ID_OK:
            return

        dcmdir = dlg.GetPath()

    # 2. load metadata about all data series in
    #    the DICOM directory. This is performed
    #    on a separate thread via the
    #    progress.runWithBounce function.
    series = []
    images = []

    def scan():
        try:
            series.extend(fsldcm.scanDir(dcmdir))
            if len(series) == 0:
                raise Exception('Could not find any DICOM '
                                'data series in {}'.format(dcmdir))

        except Exception as e:
            series.append(e)

    # 3. ask user which data series
    #    they want to load. This is called
    #    after the scan function has
    #    finished - see runWithBounce.
    def postScan(completed):

        # did the user cancel the progress dialog?
        if not completed:
            return

        # did an error occur in the scan step above?
        if isinstance(series[0], Exception):
            errTitle = strings.titles[  'loadDicom.scanError']
            errMsg   = strings.messages['loadDicom.scanError']
            status.reportError(errTitle, errMsg, series[0])
            return

        dlg = BrowseDicomDialog(parent, series)
        dlg.CentreOnParent()

        if dlg.ShowModal() != wx.ID_OK:
            return

        # load the selected series - this is
        # done asynchronously via another
        # call to runWithBounce.
        for i in reversed(list(range(len(series)))):
            if not dlg.IsSelected(i):
                series.pop(i)

        title = strings.titles[  'loadDicom.loading']
        msg   = strings.messages['loadDicom.loading']
        progress.runWithBounce(load, title, msg, callback=postLoad)

    # 4. Load the selected series. This is run
    #    on a separate thread via runWithBounce
    def load():
        try:
            for s in series:
                images.extend(fsldcm.loadSeries(s))

            if len(images) == 0:
                raise Exception('No images could be loaded '
                                'from {}'.format(dcmdir))

        except Exception as e:
            images.insert(0, e)

    # Pass the loaded images to the calback
    # function. This is called after the
    # load function has finished.
    def postLoad(completed):

        # Did the user cancel the progress dialog?
        if not completed:
            return

        # Did an error occur in the load step above?
        if isinstance(images[0], Exception):
            errTitle = strings.titles[  'loadDicom.loadError']
            errMsg   = strings.messages['loadDicom.loadError']
            status.reportError(errTitle, errMsg, images[0])
            return

        fslsettings.write('loadSaveOverlayDir',
                          op.dirname(dcmdir.rstrip(op.sep)))

        if callback is not None:
            callback(images)

    # Kick off the process
    title = strings.titles[  'loadDicom.scanning']
    msg   = strings.messages['loadDicom.scanning']
    progress.runWithBounce(scan, title, msg, callback=postScan)


class BrowseDicomDialog(wx.Dialog):
    """The ``BrowseDicomDialog`` contains a ``BrowseDicomPanel``, and a
    couple of buttons, allowing the user to select which DICOM series
    they would like to load.
    """

    def __init__(self, parent, dcmseries):
        """Create a ``BrowseDicomDialog``.

        :arg parent:    ``wx`` parent object
        :arg dcmseries: List of DICOM data series, as returned by the
                        :func:`fsl.data.dicom.scanDir` function.
        """

        wx.Dialog.__init__(self,
                           parent,
                           title=strings.titles[self],
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.__browser = BrowseDicomPanel(self, dcmseries)
        self.__load    = wx.Button(self, id=wx.ID_OK)
        self.__cancel  = wx.Button(self, id=wx.ID_CANCEL)

        self.__load  .SetDefault()
        self.__load  .SetLabel(strings.labels[self, 'load'])
        self.__cancel.SetLabel(strings.labels[self, 'cancel'])

        self.__sizer    = wx.BoxSizer(wx.VERTICAL)
        self.__btnSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND, proportion=1)
        self.__btnSizer.Add(self.__load,   flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND)
        self.__btnSizer.Add(self.__cancel, flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND)

        self.__sizer.Add((1, 10),         flag=wx.EXPAND)
        self.__sizer.Add(self.__browser,  flag=wx.EXPAND, proportion=1)
        self.__sizer.Add((1, 10),         flag=wx.EXPAND)
        self.__sizer.Add(self.__btnSizer, flag=wx.EXPAND)
        self.__sizer.Add((1, 10),         flag=wx.EXPAND)

        self.SetSizer(self.__sizer)
        self.Layout()
        self.Fit()

        self.__load  .Bind(wx.EVT_BUTTON, self.__onLoad)
        self.__cancel.Bind(wx.EVT_BUTTON, self.__onCancel)


    def __onLoad(self, ev):
        """Called when the *Load* button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_OK)


    def __onCancel(self, ev):
        """Called when the *Cancel* button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_CANCEL)


    def IsSelected(self, sidx):
        """Returns ``True`` if the DICOM series at the given index has
        been selected by the user, ``False`` otherwise.
        """
        return self.__browser.IsSelected(sidx)



class BrowseDicomPanel(wx.Panel):
    """The ``BrowseDicomPanel`` displayes information about a collection of
    DICOM data series, and allows the user to select which series they would
    like to load.
    """


    def __init__(self, parent, dcmseries):
        """Create a ``BrowseDicomPanel``.

        :arg parent:    ``wx`` parent object
        :arg dcmseries: List of DICOM data series, as returned by the
                        :func:`fsl.data.dicom.scanDir` function.
        """

        wx.Panel.__init__(self, parent)

        # we assume that this metadata
        # is the same across all series
        date        = dcmseries[0].get('AcquisitionDateTime', '')
        dcmdir      = dcmseries[0].get('DicomDir',            '')
        patient     = dcmseries[0].get('PatientName',         '')
        institution = dcmseries[0].get('InstitutionName',     '')

        try:
            date = datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%f')
            date = '{:4d}-{:2d}-{:2d}'.format(date.year, date.month, date.day)
        except ValueError:
            date = ''

        self.__dcmdirLabel      = wx.StaticText(self)
        self.__dateLabel        = wx.StaticText(self)
        self.__patientLabel     = wx.StaticText(self)
        self.__institutionLabel = wx.StaticText(self)

        self.__dcmdir      = wx.StaticText(self, style=wx.ST_ELLIPSIZE_START)
        self.__date        = wx.StaticText(self)
        self.__patient     = wx.StaticText(self)
        self.__institution = wx.StaticText(self)
        self.__series      = wg.WidgetGrid(self, style=wx.VSCROLL)

        self.__series.SetMinSize((-1, 200))

        self.__loadCheckboxes = [wx.CheckBox(self) for s in dcmseries]

        self.__dcmdirLabel     .SetLabel(strings.labels[self, 'dicomdir'])
        self.__dateLabel       .SetLabel(strings.labels[self, 'date'])
        self.__patientLabel    .SetLabel(strings.labels[self, 'patient'])
        self.__institutionLabel.SetLabel(strings.labels[self, 'institution'])
        self.__dcmdir          .SetLabel(dcmdir)
        self.__date            .SetLabel(date)
        self.__patient         .SetLabel(patient)
        self.__institution     .SetLabel(institution)

        self.__mainSizer   = wx.BoxSizer(wx.VERTICAL)
        self.__titleSizer  = wx.FlexGridSizer(2, 5, 5)
        self.__titleSizer.AddGrowableCol(1)

        self.__titleSizer.Add(self.__dcmdirLabel, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__dcmdir, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__dateLabel, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__date, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__patientLabel, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__patient, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__institutionLabel, flag=wx.EXPAND)
        self.__titleSizer.Add(self.__institution, flag=wx.EXPAND)

        self.__mainSizer.Add(self.__titleSizer,
                             flag=wx.EXPAND | wx.ALL,
                             border=5)
        self.__mainSizer.Add(self.__series,
                             flag=wx.EXPAND | wx.ALL,
                             border=5,
                             proportion=1)

        self.SetSizer(self.__mainSizer)

        # columns:
        #   SeriesNumber[.EchoNumber]
        #   SeriesDescription
        #   ReconMatrix
        #   Load (checkbox)
        # TODO For other useful information,
        #      you might need to look in the niftis

        # set up the grid
        self.__series.SetGridSize(len(dcmseries), 4, growCols=(0, 1))
        self.__series.ShowColLabels()
        self.__series.SetColLabel(0, strings.labels[self, 'SeriesNumber'])
        self.__series.SetColLabel(1, strings.labels[self, 'SeriesDescription'])
        self.__series.SetColLabel(2, strings.labels[self, 'Matrix'])
        self.__series.SetColLabel(3, strings.labels[self, 'Load'])

        for i, s in enumerate(dcmseries):

            num  = str(s.get('SeriesNumber',      '?'))
            echo = str(s.get('EchoNumber',        ''))
            desc =     s.get('SeriesDescription', 'n/a')
            size =     s.get('ReconMatrixPE',     'n/a')

            if len(echo) > 0:
                echo = '.' + echo

            self.__series.SetText(  i, 0, num + echo)
            self.__series.SetText(  i, 1, desc)
            self.__series.SetText(  i, 2, '{}x{}'.format(size, size))
            self.__series.SetWidget(i, 3, self.__loadCheckboxes[i])

        self.__series.Refresh()


    def IsSelected(self, sidx):
        """Returns ``True`` if the DICOM series at the given index has
        been selected by the user, ``False`` otherwise.
        """
        return self.__loadCheckboxes[sidx].GetValue()



#!/usr/bin/env python
#
# lightboxviewprofile.py - The LightBoxViewProfile class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LightBoxViewProfile` class, an interaction
:class:`.Profile` for :class:`.LightBoxPanel` views.
"""

import logging

import wx

import fsleyes.profiles as profiles
import fsl.utils.idle   as idle


log = logging.getLogger(__name__)


class LightBoxViewProfile(profiles.Profile):
    """The ``LightBoxViewProfile`` is an interaction profile for
    :class:`.LightBoxPanel` views. It defines mouse/keyboard handlers which
    allow the user to navigate through the ``LightBoxPanel`` display of the
    overlays in the :class:`.OverlayList`.

    ``LightBoxViewProfile`` defines two *modes* (see the :class:`.Profile`
    class documentation):

    ======== ==================================================================
    ``view`` The user can change the :attr:`.DisplayContext.location` via
             left mouse drags, and can change the
             :attr:`.LightBoxCanvasOpts.topRow` via the mouse wheel.

    ``zoom`` The user can change the :attr:`.LightBoxCanvasOpts.ncols` property
             with the mouse wheel (effectively zooming in/out of the canvas).
    ======== ==================================================================
    """


    @staticmethod
    def supportedView():
        """Returns the :class:`.LightBoxPanel` class. """
        import fsleyes.views.lightboxpanel as lightboxpanel
        return lightboxpanel.LightBoxPanel


    @staticmethod
    def tempModes():
        """Returns the temporary mode map for the ``LightBoxViewProfile``,
        which controls the use of modifier keys to temporarily enter other
        interaction modes.
        """
        # Command/CTRL puts the user in zoom mode,
        # and ALT puts the user in pan mode
        return {('view', wx.WXK_CONTROL) : 'zoom'}


    @staticmethod
    def altHandlers():
        """Returns the alternate handlers map, which allows event handlers
        defined in one mode to be re-used whilst in another mode.
        """
        return { ('view', 'LeftMouseDown') :('view', 'LeftMouseDrag')}


    def __init__(self, viewPanel, overlayList, displayCtx):
        """Create a ``LightBoxViewProfile``.

        :arg viewPanel:    A :class:`.LightBoxPanel` instance.
        :arg overlayList:  The :class:`.OverlayList` instance.
        :arg displayCtx:   The :class:`.DisplayContext` instance.
        """

        profiles.Profile.__init__(self,
                                  viewPanel,
                                  overlayList,
                                  displayCtx,
                                  modes=['view', 'zoom'])

        self.__canvas = viewPanel.getCanvas()


    def getEventTargets(self):
        """Returns the :class:`.LightBoxCanvas` contained in the
        :class:`.LightBoxPanel`, which is the target for all mouse/keyboard
        events.
        """
        return [self.__canvas]


    def _viewModeMouseWheel(self,
                            ev,
                            canvas,
                            wheel,
                            mousePos=None,
                            canvasPos=None):
        """Handles mouse wheel events in ``view`` mode.

        Updates the :attr:.LightBoxCanvasOpts.topRow` property, thus scrolling
        through the slices displayed on the canvas.
        """

        # When we scroll up, we move to lower slices,
        # so a positive scroll direction corresponds
        # to negative slice direction, and vice versa.
        if   wheel > 0: wheel = -1
        elif wheel < 0: wheel =  1
        else:           return False

        opts = self.__canvas.opts

        # See comment in OrthoViewProfile._zoomModeMouseWheel
        # about timeout
        def update():
            opts.topRow += wheel

        idle.idle(update, timeout=0.1)

        return True


    def _viewModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse drags in ``view`` mode.

        Updates the :attr:`.DisplayContext.location` property to track the
        mouse location.
        """

        if canvasPos is None:
            return False

        self.displayCtx.location.xyz = canvasPos

        return True


    def _zoomModeMouseWheel(self,
                            ev,
                            canvas,
                            wheel,
                            mousePos=None,
                            canvasPos=None):
        """Handles mouse wheel events in ``zoom`` mode.

        Zooms in/out of the canvas by updating the :attr:`.SceneOpts.zoom`
        property.
        """

        if   wheel > 0: wheel =  50
        elif wheel < 0: wheel = -50
        else:           return False

        opts = self.viewPanel.sceneOpts

        # see comment in OrthoViewProfile._zoomModeMouseWheel
        # about timeout
        def update():
            opts.zoom += wheel

        idle.idle(update, timeout=0.1)

        return True



#!/usr/bin/env python
#
# query.py - The FileTreeQuery class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
# Author: Michiel Cottaar <michiel.cottaar@.ndcn.ox.ac.uk>
#
"""This module contains the :class:`FileTreeQuery` class, which can be used to
search for files in a directory described by a `FileTree
<https://git.fmrib.ox.ac.uk/ndcn0236/file-tree/>`_. A ``FileTreeQuery`` object
returns :class:`Match` objects which each represent a file that is described
by the ``FileTree``, and which is present in the directory.

The following utility functions, used by the ``FileTreeQuery`` class, are also
defined in this module:

.. autosummary::
   :nosignatures:

   scan
"""


import              logging
import              collections
import functools as ft

import numpy     as np


log = logging.getLogger(__name__)


class FileTreeQuery:
    """The ``FileTreeQuery`` class uses a ``FileTree`` to search
    a directory for files which match a specific query.

    A ``FileTreeQuery`` scans the contents of a directory which is described
    by a ``FileTree``, and identifies all file types (a.k.a. *templates*
    or *short names*) that are present, and the values of variables within each
    short name that are present. The :meth:`query` method can be used to
    retrieve files which match a specific template, and variable values.

    The :meth:`query` method returns a collection of :class:`Match` objects,
    each of which represents one file which matches the query.

    Example usage::

        >>> from file_tree import FileTree
        >>> from fsleyes.filetree import FileTreeQuery

        >>> tree  = FileTree.read('bids_raw', './my_bids_data')
        >>> query = FileTreeQuery(tree)

        >>> query.axes('anat_image')
        ['acq', 'ext', 'modality', 'participant', 'rec', 'run_index',
         'session']

        >>> query.variables('anat_image')
        {'acq': [None],
         'ext': ['.nii.gz'],
         'modality': ['T1w', 'T2w'],
         'participant': ['01', '02', '03'],
         'rec': [None],
         'run_index': [None, '01', '02', '03'],
         'session': [None]}

        >>> query.query('anat_image', participant='01')
        [Match(./my_bids_data/sub-01/anat/sub-01_T1w.nii.gz),
         Match(./my_bids_data/sub-01/anat/sub-01_T2w.nii.gz)]


    Matches for templates contained within sub-trees are referred to by
    constructing a hierarchical path from the sub-tree template name(s),
    and the template name - see the :meth:`Match.full_name` method.
    """


    def __init__(self, tree):
        """Create a ``FileTreeQuery``. The contents of the tree directory are
        scanned via the :func:`scan` function, which may take some time for
        large data sets.

        :arg tree: The ``FileTree`` object
        """
        self.__tree        = tree
        self.__matcharrays = scan(tree)


    def axes(self, template):
        """Returns a list containing the names of variables present in files
        of the given ``template`` type.
        """
        return list(self.__matcharrays[template].coords.keys())


    def variables(self, template=None):
        """Return a dict of ``{variable : [values]}`` mappings.
        This dict describes all variables and their possible values in
        the tree.

        If a ``template`` is specified, only variables which are present in
        files of that ``template`` type are returned.
        """
        if template is not None:
            templates = [template]
        else:
            templates = self.__matcharrays.keys()

        variables = collections.defaultdict(set)

        for template in templates:
            coords = self.__matcharrays[template].coords

            for axis in coords.keys():
                varvalues       = variables[axis]
                variables[axis] = varvalues.union(set(coords[axis].data))

        # Variable values will usually be strings,
        # but can sometimes be None, so we convert
        # to str to handle this.
        variables = {name : sorted(vals, key=str)
                     for name, vals in variables.items()}

        return variables


    @property
    def tree(self):
        """Returns the ``FileTree`` associated with this ``FileTreeQuery``.
        """
        return self.__tree


    @property
    def templates(self):
        """Returns a list containing all templates of the ``FileTree`` that
        are present in the directory.
        """
        return list(self.__matcharrays.keys())


    def matcharray(self, template):
        """Returns a reference to the ``xarray.DataArray`` which contains the
        file paths for the given ``template``.
        """
        return self.__matcharrays[template]


    def query(self, template, **variables):
        """Search for files of the given ``template``, which match
        the specified ``variables``. All hits are returned for variables
        that are unspecified.

        :arg template: Template of files to search for.

        All other arguments are assumed to be ``variable=value`` pairs,
        used to restrict which matches are returned. All values are returned
        for variables that are not specified, or variables which are given a
        value of ``'*'``.

        :returns: A list  of ``Match`` objects
        """

        # Build a slice containing a value for
        # every axis of the template array
        varnames    = list(variables.keys())
        allvarnames = self.variables(template).keys()
        matcharray  = self.__matcharrays[ template]
        slc         = []

        for var in allvarnames:

            if var in varnames: val = variables[var]
            else:               val = '*'

            if val == '*': slc.append(slice(None))
            else:          slc.append(val)

        # Retrieve the results
        results = matcharray.loc[tuple(slc)]

        # Convert xarray.DataArray into a list of
        # Match objects. I can't find an elegant
        # way to do this - something like apply_ufunc
        # would be nice, but we lose the labelling
        # information.
        matches = []
        riter   = np.nditer(results, flags=['multi_index'])

        for fname in riter:

            fname = fname.item()

            if fname == '':
                continue

            # Look up the variable values associated
            # with this file name, and create a
            # corresponding Match object
            index  = riter.multi_index
            coords = results[index].coords
            rvars  = {ax : coords[ax].data[()] for ax in coords}

            matches.append(Match(fname, template, rvars))

        return matches


@ft.total_ordering
class Match:
    """A ``Match`` object represents a file with a name matching a template in
    a ``FileTree``.  The :meth:`FileTree.query` method returns ``Match``
    objects.
    """


    def __init__(self, filename, template, variables):
        """Create a ``Match`` object. All arguments are added as attributes.

        :arg template:   template identifier
        :arg value:

        """
        self.__filename  = filename
        self.__template  = template
        self.__variables = variables


    @property
    def filename(self):
        return self.__filename


    @property
    def template(self):
        return self.__template


    @property
    def variables(self):
        return dict(self.__variables)


    def __eq__(self, other):
        return (isinstance(other, Match)            and
                self.filename   == other.filename   and
                self.template   == other.template   and
                self.variables  == other.variables)


    def __lt__(self, other):
        return isinstance(other, Match) and self.filename < other.filename


    def __repr__(self):
        """Returns a string representation of this ``Match``. """
        return 'Match({}: {})'.format(self.template, self.filename)


    def __str__(self):
        """Returns a string representation of this ``Match``. """
        return repr(self)


def scan(tree, filterEmpty=True):
    """Scans the directory of the given ``FileTree`` to find all files which
    match a tree template.

    A dictionary of ``{template : xarray.DataArray}`` mappings is returned,
    where each ``DataArray`` has dimensions corresponding to variables used in
    the template, and contains names (as strings) of matching files present on
    disk as strings. Entries in an arrays for variable values which do not
    exist on disk are set to the empty string.

    See the ``file_tree.FileTree.get_mult_glob`` method for more details.

    :arg tree:        ``FileTree`` to scan

    :arg filterEmpty: If ``True`` (the default), file tree templates which
                      do not match any files on disk are not returned.

    :returns:         A dict of ``{template : xarray.DataArray}`` objects,
                      one for each template.
    """

    templates = tree.template_keys(only_leaves=True)
    xarrays   = tree.get_mult_glob(templates)
    results   = {}

    for template in templates:

        xa = xarrays[template]

        # Skip templates which do not have
        # any files present on disk
        if filterEmpty:
            if (xa == '').sum() == np.prod(xa.shape):
                continue

        results[template] = xa

    return results



#!/usr/bin/env python
#
# orthoviewprofile.py - The OrthoViewProfile class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoViewProfile` class, an interaction
:class:`.Profile` for :class:`.OrthoPanel` views.
"""

import logging

import wx

import numpy as np

import fsl.utils.idle       as idle
import fsl.transform.affine as affine

import fsleyes.profiles     as profiles
import fsleyes.actions      as actions


log = logging.getLogger(__name__)


class OrthoViewProfile(profiles.Profile):
    """The ``OrthoViewProfile`` class is a :class:`.Profile` for the
    :class:`.OrthoPanel` class.  It defines mouse/keyboard handlers which
    allow the user to navigate through the ``OrthoPanel`` display of the
    overlays in the :class:`.OverlayList`.

    ``OrthoViewProfile`` defines the following *modes* (see the
    :class:`.Profile` class documentation):

    ================ ========================================================
    ``nav``          The user can change the currently displayed location.
                     This is accomplished by updating the
                     :attr:`.DisplayContext.location` property on left mouse
                     drags.

    ``slice``        The user can change the current slice shown on a single
                     canvas.

    ``zoom``         The user can zoom in/out of a canvas with the mouse
                     wheel, and draw a rectangle on a canvas in which to
                     zoom. This is accomplished by updating the
                     :attr:`.SliceCanvasOpts.zoom` property on mouse wheel
                     changes, and displaying a :class:`~.annotations.Rect`
                     annotation on left mouse drags.

    ``pan``          The user can pan around a canvas (if the canvas is
                     zoomed in). This is accomplished by calling the
                     :meth:`.SliceCanvas.panDisplayBy` on left mouse drags.

    ``bricon``       The user can drag the mouse along a canvas to change
                     the brightness/contrast of the currently selected
                     overlay.

    ``regionBricon`` The user can draw a rectangle on a canvas; the minimum /
                     maximum display range of the currently selected overlay
                     (assuming it is an :class:`.Image` being displayed as
                     a ``volume``) is set to the minimum/maxmimum intensity
                     within the region.

    ``pick``         If the currently selected overlay is a :class:`.Mesh`,
                     the user can select the mesh vertex which is nearest to
                     the mouse click.
    ================ ========================================================


    The ``OrthoViewProfile`` class also defines a few :mod:`.actions`:

    .. autosummary::
       :nosignatures:

       resetDisplay
       centreCursor
    """


    @staticmethod
    def supportedView():
        """Returns the :class:`.OrthoPanel` class. """
        import fsleyes.views.orthopanel as orthopanel
        return orthopanel.OrthoPanel


    @staticmethod
    def tempModes():
        """Returns the temporary mode map for the ``OrthoViewProfile``,
        which controls the use of modifier keys to temporarily enter other
        interaction modes.
        """
        # Command/CTRL puts the user in zoom mode,
        # and ALT puts the user in pan mode
        return {
            ('nav',  wx.WXK_CONTROL)                : 'zoom',
            ('nav',  wx.WXK_ALT)                    : 'pan',
            ('nav',  wx.WXK_SHIFT)                  : 'slice',
            ('nav', (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'bricon'}


    @staticmethod
    def altHandlers():
        """Returns the alternate handlers map, which allows event handlers
        defined in one mode to be re-used whilst in another mode.
        """
        return {
            # in navigate, slice, and zoom mode, the
            # left mouse button navigates, the right
            # mouse button draws a zoom rectangle,
            # and the middle button pans.
            ('nav',  'LeftMouseDown')   : ('nav',  'LeftMouseDrag'),
            ('nav',  'MiddleMouseDrag') : ('pan',  'LeftMouseDrag'),
            ('nav',  'RightMouseDown')  : ('zoom', 'RightMouseDown'),
            ('nav',  'RightMouseDrag')  : ('zoom', 'RightMouseDrag'),
            ('nav',  'RightMouseUp')    : ('zoom', 'RightMouseUp'),

            # In slice mode, the left and right
            # mouse buttons work as for pick mode.
            # Middle and right mouse works the
            # same as for nav mode, defined above.
            ('slice', 'LeftMouseDown')   : ('pick', 'LeftMouseDown'),
            ('slice', 'LeftMouseDrag')   : ('pick', 'LeftMouseDrag'),
            ('slice', 'MiddleMouseDrag') : ('pan',  'LeftMouseDrag'),
            ('slice', 'RightMouseDown')  : ('zoom', 'RightMouseDown'),
            ('slice', 'RightMouseDrag')  : ('zoom', 'RightMouseDrag'),
            ('slice', 'RightMouseUp')    : ('zoom', 'RightMouseUp'),

            # In zoom mode, the left mouse button
            # navigates, the right mouse button
            # draws a zoom rectangle, and the
            # middle mouse button pans
            ('zoom', 'RightMouseDown')  : ('zoom', 'RightMouseDrag'),
            ('zoom', 'LeftMouseDown')   : ('nav',  'LeftMouseDown'),
            ('zoom', 'LeftMouseDrag')   : ('nav',  'LeftMouseDrag'),
            ('zoom', 'MiddleMouseDrag') : ('pan',  'LeftMouseDrag'),

            # In pick mode, left mouse down
            # is the same as left mouse drag.
            ('pick', 'LeftMouseDown') : ('pick', 'LeftMouseDrag'),

            # In bricon mode, right click+drag
            # invokes regionBricon mode
            ('bricon', 'RightMouseDown') : ('regionBricon', 'LeftMouseDrag'),
            ('bricon', 'RightMouseDrag') : ('regionBricon', 'LeftMouseDrag'),
            ('bricon', 'RightMouseUp')   : ('regionBricon', 'LeftMouseUp')}


    @staticmethod
    def fallbackHandlers():
        """Returns the fallback handler map, which allows an event to be
        forwarded to a second handler if the primary handler returns ``False``.
        """
        return {
            ('pick', 'LeftMouseDown') : ('nav', 'LeftMouseDown'),
            ('pick', 'LeftMouseDrag') : ('nav', 'LeftMouseDrag')}


    def __init__(self,
                 viewPanel,
                 overlayList,
                 displayCtx,
                 extraModes=None):
        """Creates an :class:`OrthoViewProfile`, which can be registered
        with the given ``viewPanel``.


        .. note:: The :class:`.OrthoEditProfile` is a sub-class of the
                  ``OrthoViewProfile``. It uses the ``extraModes`` argument
                  to set up its edit-related modes.


        :arg viewPanel:    An :class:`.OrthoPanel` instance.

        :arg overlayList:  The :class:`.OverlayList` instance.

        :arg displayCtx:   The :class:`.DisplayContext` instance.

        :arg extraModes:   Extra modes to pass through to the
                           :class:`.Profile` constructor.
        """

        if extraModes is None:
            extraModes = []

        modes = ['nav', 'slice', 'pan', 'zoom', 'bricon',
                 'pick', 'regionBricon'] + extraModes

        profiles.Profile.__init__(self,
                                  viewPanel,
                                  overlayList,
                                  displayCtx,
                                  modes)

        # We create our own name and use it
        # for registering property listeners,
        # so sub-classes can use Profile.name
        # to register its own listeners.
        self.__name    = 'OrthoViewProfile_{}'.format(self.name)

        self.__xcanvas = viewPanel.getXCanvas()
        self.__ycanvas = viewPanel.getYCanvas()
        self.__zcanvas = viewPanel.getZCanvas()

        # These attributes will occasionally store
        # references to gl.annotations.Rect objects -
        # see the _zoomModeRightMouse*  and
        # _regionBriconModeLeftMouse* handlers
        self.__zoomLastRect         = None
        self.__regionBriconLastRect = None

        # This attribute is used by the
        # bricon mode methods
        self.__briconOrigin = None

        overlayList.addListener('overlays',
                                self.__name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__selectedOverlayChanged)


    def destroy(self):
        """Must be called when this ``OrthoViewProfile`` is no longer needed.
        Removes some property listeners, and calls :meth:`.Profile.destroy`.
        """
        self.overlayList.removeListener('overlays',        self.__name)
        self.displayCtx .removeListener('selectedOverlay', self.__name)
        profiles.Profile.destroy(self)


    def getEventTargets(self):
        """Overrides :meth:`.Profile.getEventTargets`.

        Returns the three :class:`.SliceCanvas` instances displayed in the
        :class:`.OrthoPanel` instance that is using this ``OrthoViewProfile``.
        """
        return [self.__xcanvas, self.__ycanvas, self.__zcanvas]


    def __selectedOverlayChanged(self, *a):
        """Called when the :class:`.OverlayList` or
        :attr:`.DisplayContext.selectedOverlay` changes. Enables/disables
        the action methods based on the newly selected overlay.
        """

        ovl = self.displayCtx.getSelectedOverlay()

        self.resetDisplay     .enabled = ovl is not None
        self.centreCursor     .enabled = ovl is not None
        self.centreCursorWorld.enabled = ovl is not None


    @actions.action
    def resetDisplay(self):
        """Resets the display bounds for all canvases. Also resets the
        :class:`.SceneOpts.zoom`, :class:`.OrthoOpts.xzoom`,
        :class:`.OrthoOpts.yzoom`, and :class:`.OrthoOpts.zzoom` properties to
        100%.

        See the :meth:`.SliceCanvas.resetDisplay` method.
        """

        opts = self.viewPanel.sceneOpts

        self.__xcanvas.resetDisplay()
        self.__ycanvas.resetDisplay()
        self.__zcanvas.resetDisplay()

        opts.zoom  = 100
        opts.xzoom = 100
        opts.yzoom = 100
        opts.zzoom = 100


    @actions.action
    def centreCursor(self):
        """Sets the :attr:`.DisplayContext.location` to the centre of the
        :attr:`.DisplayContext.bounds`.
        """

        bounds = self.displayCtx.bounds

        xmid = bounds.xlo + 0.5 * bounds.xlen
        ymid = bounds.ylo + 0.5 * bounds.ylen
        zmid = bounds.zlo + 0.5 * bounds.zlen

        self.displayCtx.location.xyz = [xmid, ymid, zmid]


    @actions.action
    def centreCursorWorld(self):
        """Sets the :attr:`.DisplayContext.location` to world location
        (0, 0, 0), where the 'world' is in terms of the currently selected
        ovelray.
        """

        ovl    = self.displayCtx.getSelectedOverlay()
        refOvl = self.displayCtx.getReferenceImage(ovl)

        if refOvl is not None:
            opts   = self.displayCtx.getOpts(refOvl)
            origin = opts.transformCoords([0, 0, 0], 'world', 'display')
        else:
            origin = [0, 0, 0]

        self.displayCtx.location.xyz = origin


    ########################
    # Navigate mode handlers
    ########################


    def __offsetLocation(self, x, y, z):
        """Used by some ``nav`` mode handlers. Returns a sequence of three
        values, one per display space axis, which specify the amount by
        which the :attr:`.DisplayContext.location` should be changed,
        according to the directions specified by the ``x``, ``y``, and ``z``
        arguments.

        If the currently selected overlay is an :class:`.Nifti` instance, the
        distance that a navigation operation should shift the display will
        differ depending on the value of the :attr:`.NiftiOpts.transform`
        property. For example, if ``transform`` is ``id``, the display should
        be moved by one unit (which corresponds to one voxel). But if the
        ``transform`` is ``pixdim``, the display should be moved by one pixdim
        (e.g. 2, for a "math:`2mm^3` image).

        Each of the ``x``, ``y`` and ``z`` arguments are interpreted as being
        positive, zero, or negative. A positive/negative value indicates that
        the display location should be increased/decreased along the
        corresponding axis, and zero indicates that the location should stay
        the same along an axis.
        """

        overlay = self.displayCtx.getReferenceImage(
            self.displayCtx.getSelectedOverlay())

        # If non-zero, round to -1 or +1
        x = np.sign(x) * np.ceil(np.abs(np.clip(x, -1, 1)))
        y = np.sign(y) * np.ceil(np.abs(np.clip(y, -1, 1)))
        z = np.sign(z) * np.ceil(np.abs(np.clip(z, -1, 1)))

        # If the currently selected overlay
        # is non-volumetric, and does not
        # have a reference image, we'll just
        # move by +/-1 along each axis (as
        # specified by the x/y/z parameters).
        if overlay is None:
            dloc     = self.displayCtx.location.xyz
            dloc[0] += x
            dloc[1] += y
            dloc[2] += z

        # But if we have a voluemtric reference
        # image to play with, we're going to
        # move to the next/previous voxel along
        # each axis (as specified by x/y/z),
        # and calculate the corresponding location
        # in display space coordinates.
        else:

            # We use this complicated looking
            # code so that the adjusted location
            # is centered within the next/previous
            # voxel on the depth axis.
            #
            # The procedure is as follows:
            #
            #   1. Calculate the current display
            #      location in voxels. If we are
            #      displaying in id/pixdim space,
            #      we round the voxels to integers,
            #      otherwise we use floating point
            #      voxel coordinates.
            #
            #   2. Offset the voxel coordinates
            #      according to the x/y/z parameters. To
            #      do this we use the Image.axisMapping
            #      method which returns the approximate
            #      correspondence between voxel axes and
            #      display axes.
            #
            #   3. Transform the voxel coordinates back
            #      into the display coordinate system.

            offsets  = [x, y, z]
            opts     = self.displayCtx.getOpts(overlay)
            vround   = opts.transform in ('id', 'pixdim', 'pixdim-flip')
            vloc     = opts.getVoxel(clip=False, vround=vround)
            voxAxes  = overlay.axisMapping(opts.getTransform('voxel',
                                                             'display'))

            for i in range(3):
                vdir       = np.sign(voxAxes[i])
                vax        = np.abs(voxAxes[i]) - 1
                vloc[vax] += vdir * offsets[i]

            dloc = opts.transformCoords([vloc], 'voxel', 'display')[0]

        return dloc


    def _navModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse drags in ``nav`` mode.

        Left mouse drags in ``nav`` mode update the
        :attr:`.DisplayContext.location` to follow the mouse location.
        """

        if canvasPos is None:
            return False

        self.displayCtx.location = canvasPos

        return True


    def _navModeChar(self, ev, canvas, key):
        """Handles key presses in ``nav`` mode.

        Arrow key presses in ``nav`` mode update the
        :attr:`.DisplayContext.location`.  Arrow keys map to the
        horizontal/vertical axes, and -/+ keys map to the depth axis of the
        canvas which was the target of the event.
        """

        if len(self.overlayList) == 0:
            return False

        try:              ch = chr(key)
        except Exception: ch = None

        dirs  = [0, 0, 0]
        copts = canvas.opts

        if   key == wx.WXK_LEFT:  dirs[copts.xax] = -1
        elif key == wx.WXK_RIGHT: dirs[copts.xax] =  1
        elif key == wx.WXK_UP:    dirs[copts.yax] =  1
        elif key == wx.WXK_DOWN:  dirs[copts.yax] = -1
        elif ch  in ('+', '='):   dirs[copts.zax] =  1
        elif ch  in ('-', '_'):   dirs[copts.zax] = -1
        else:                     return False

        def update():
            self.displayCtx.location.xyz = self.__offsetLocation(*dirs)

        # See comment in _zoomModeMouseWheel about timeout
        idle.idle(update, timeout=0.1)

        return True


    #####################
    # Slice mode handlers
    #####################


    def _sliceModeMouseWheel(
            self, ev, canvas, wheel, mousePos=None, canvasPos=None):
        """Handles mouse wheel movement in ``slice`` mode.

        Mouse wheel movement on a canvas changes the depth location displayed
        on that canvas.
        """

        if len(self.overlayList) == 0:
            return False

        dirs  = [0, 0, 0]
        copts = canvas.opts

        if   wheel > 0: dirs[copts.zax] =  1
        elif wheel < 0: dirs[copts.zax] = -1

        pos = self.__offsetLocation(*dirs)

        def update():
            self.displayCtx.location[copts.zax] = pos[copts.zax]

        # See comment in _zoomModeMouseWheel about timeout
        idle.idle(update, timeout=0.1)

        return True


    def _sliceModeChar(self, ev, canvas, key):
        """Handles character events in ``slice`` mode.

        Up/down and Page-up/Page-down keys will increase/decrease the
        depth location on the current canvas.
        """

        if   key in (wx.WXK_UP,   wx.WXK_PAGEUP):   off =  1
        elif key in (wx.WXK_DOWN, wx.WXK_PAGEDOWN): off = -1
        else:                                       return False

        return self._sliceModeMouseWheel(ev, canvas, off)


    ####################
    # Zoom mode handlers
    ####################


    def _zoomModeMouseWheel(self,
                            ev,
                            canvas,
                            wheel,
                            mousePos=None,
                            canvasPos=None):
        """Handles mouse wheel events in ``zoom`` mode.

        Mouse wheel motion in zoom mode increases/decreases the zoom level
        of the target canvas.
        """
        if   wheel > 0: wheel =  50
        elif wheel < 0: wheel = -50

        copts   = canvas.opts
        minzoom = copts.getAttribute('zoom', 'minval')
        maxzoom = copts.getAttribute('zoom', 'maxval')

        # Over SSH/X11, mouse wheel events seem to get queued,
        # and continue to get processed after the user has
        # stopped spinning the mouse wheel, which is super
        # frustrating. So we do the update asynchronously, and
        # set a time out to drop the event, and prevent the
        # horribleness from happening.
        def update():
            newZoom    = np.clip(copts.zoom + wheel, minzoom, maxzoom)
            copts.zoom = newZoom

        idle.idle(update, timeout=0.1)

        return True


    def _zoomModeChar(self, ev, canvas, key):
        """Handles key presses in ``zoom`` mode.

        The +/- keys in zoom mode increase/decrease the zoom level
        of the target canvas.
        """

        try:              ch = chr(key)
        except Exception: ch = None

        zoom = 0

        if   ch in ('-', '_'):  zoom = -1
        elif ch in ('=', '+'):  zoom =  1
        else:                   return False

        return self._zoomModeMouseWheel(None, canvas, zoom)


    def _zoomModeRightMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles right mouse drags in ``zoom`` mode.

        Right mouse drags in zoom mode draw a rectangle on the target
        canvas.

        When the user releases the mouse (see :meth:`_zoomModeLeftMouseUp`),
        the canvas will be zoomed in to the drawn rectangle.
        """

        mouseDownPos, canvasDownPos = self.getMouseDownLocation()

        if canvasPos is None or canvasDownPos is None:
            return False

        copts  = canvas.opts
        x, y   = [canvasDownPos[copts.xax], canvasDownPos[copts.yax]]
        width  = canvasPos[copts.xax] - x
        height = canvasPos[copts.yax] - y

        self.__zoomLastRect = canvas.getAnnotations().rect(
            x, y,
            width,
            height,
            width=1,
            filled=False,
            colour=(1, 1, 0))
        canvas.Refresh()

        return True


    def _zoomModeRightMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Handles right mouse up events in ``zoom`` mode.

        When the right mouse is released in zoom mode, the target
        canvas is zoomed in to the rectangle region that was drawn by the
        user.
        """

        mouseDownPos, canvasDownPos = self.getMouseDownLocation()

        if canvasPos     is None or \
           mouseDownPos  is None or \
           canvasDownPos is None:
            return False

        if self.__zoomLastRect is not None:
            canvas.getAnnotations().dequeue(self.__zoomLastRect)
            self.__zoomLastRect = None

        copts = canvas.opts
        xlo   = min(canvasPos[copts.xax], canvasDownPos[copts.xax])
        xhi   = max(canvasPos[copts.xax], canvasDownPos[copts.xax])
        ylo   = min(canvasPos[copts.yax], canvasDownPos[copts.yax])
        yhi   = max(canvasPos[copts.yax], canvasDownPos[copts.yax])

        canvas.zoomTo(xlo, xhi, ylo, yhi)

        return True


    ###################
    # Pan mode handlers
    ###################


    def _panModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse drags in ``pan`` mode.

        Left mouse drags in pan mode move the target canvas display about
        to follow the mouse.

        If the target canvas is not zoomed in, this has no effect.
        """

        mouseDownPos, canvasDownPos = self.getMouseDownLocation()

        if canvasPos     is None: return False
        if canvasDownPos is None: return False

        copts = canvas.opts
        xoff  = canvasPos[copts.xax] - canvasDownPos[copts.xax]
        yoff  = canvasPos[copts.yax] - canvasDownPos[copts.yax]

        canvas.panDisplayBy(-xoff, -yoff)

        return True


    def _panModeChar(self, ev, canvas, key):
        """Handles key presses in ``pan`` mode.

        The arrow keys in pan mode move the target canvas display around
        (unless the canvas is not zoomed in).
        """

        xoff = 0
        yoff = 0

        if   key == wx.WXK_DOWN:  yoff = -2
        elif key == wx.WXK_UP:    yoff =  2
        elif key == wx.WXK_LEFT:  xoff = -2
        elif key == wx.WXK_RIGHT: xoff =  2
        else:                     return False

        def update():
            canvas.panDisplayBy(xoff, yoff)

        # See comment in _zoomModeMouseWheel about timeout
        idle.idle(update, timeout=0.1)

        return True


    #############
    # Bricon mode
    #############


    def _briconModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse down events in ``bricon`` mode. Stores the
        current :attr:`.Display.brightness`/:attr:`.Display.contrast` for the
        currently selected overlay.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return False

        display = self.displayCtx.getDisplay(overlay)

        self.__briconOrigin = (display.brightness, display.contrast)

        return True


    def _briconModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse up events in ``bricon`` mode. """

        self.__briconOrigin = None

        return True


    def _briconModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse drags in ``bricon`` mode.

        The brightness and contrast of the currently selected overlay are
        adjusted according to the location of the mouse, relative to the
        canvas.
        """

        if self.__briconOrigin is None:
            return False

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return False

        # o: original
        # b: brightness
        # c: contrast
        # x: mousex
        # y: mousey
        # w: canvas width
        # h: canvas height

        (ox, oy), _ = self.getMouseDownLocation()
        (ob, oc)    = self.__briconOrigin

        display = self.displayCtx.getDisplay(overlay)
        w, h    = canvas.GetSize()
        x, y    = mousePos

        # Brightness is mapped to the horizontal
        # axis, and contrast to the vertical axis.
        # They are mapped in a piece-wise linear
        # manner, where the original bricon maps
        # to the mouse down location, and 0/100
        # map to the low/high canvas edges
        # respectively.
        if x < ox: b = ob * (float(x) / ox)
        else:      b = ob + (100 - ob) * float(x - ox) / (w - ox)

        if y < oy: c = oc * (float(y) / oy)
        else:      c = oc + (100 - oc) * float(y - oy) / (h - oy)

        log.debug('Adjusting bricon for {} '
                  '(brightness: {}, contrast: {})'.format(
                      overlay.name, b, c))

        b = np.clip(b, 0, 100)
        c = np.clip(c, 0, 100)

        # Bricon are disabled when
        # VolumeOpts instances are
        # using a negative colour map.
        if display.propertyIsEnabled('brightness'): display.brightness = b
        if display.propertyIsEnabled('contrast'):   display.contrast   = c

        return True


    ###########
    # Pick mode
    ###########


    def _pickModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Handles left mouse drag events in ``pick`` mode. If the currently
        selected overlay is a :class:`.Mesh`, identifies the mesh vertex which
        is nearest to the mouse click, and updates the
        :attr:`.DisplayContext.location`. Otherwise, returns ``False``.
        """

        import fsl.data.mesh as fslmesh

        overlay = self.displayCtx.getSelectedOverlay()

        if not isinstance(overlay, fslmesh.Mesh):
            return False

        opts  = self.displayCtx.getOpts(overlay)
        loc   = opts.transformCoords(canvasPos, 'display', 'mesh')

        # The GLMesh caches the most
        # recently drawn cross section
        # vertices, and corresponding
        # mesh triangle indices, in
        # the mesh coordinate system.
        xsect = self.overlayList.getData(
            overlay,
            'crosssection_{}'.format(canvas.opts.zax), None)

        # No cross-section calculated,
        # can't pick the nearest vertex.
        if xsect is None or len(xsect[0]) == 0:
            return False

        lines, faces = xsect
        lines        = lines.reshape(-1, 3)

        # Find the location on the cross section
        # that was nearest to the mouse click
        ldists = affine.veclength(loc - lines)
        lidx   = np.argsort(ldists)[0]
        lvert  = lines[lidx]
        fidx   = faces[int(np.floor(lidx / 2))]

        # Get the triangle on the mesh
        # corresponding to this location
        face      = overlay.indices[ fidx]
        faceVerts = overlay.vertices[face]

        # Calculate the nearest vertex on
        # this triangle
        fdists = affine.veclength(faceVerts - lvert)
        vidx   = np.argsort(fdists)[0]
        vidx   = face[vidx]
        loc    = overlay.vertices[vidx, :]
        loc    = opts.transformCoords(loc, 'mesh', 'display')

        self.displayCtx.location.xyz = loc


    ###################
    # regionBricon mode
    ###################


    def _regionBriconModeLeftMouseDrag(
            self, ev, canvas, mousePos, canvasPos):
        """Handles mouse drags in ``regionBricon`` mode.

        Right mouse drags in this mode draw a rectangle on the target
        canvas. When the user releases the mouse (see
        :meth:`_regionBriconModeLeftMouseUp`), the min/max display range
        of the currently selected overlay (assuming it is a ``volume``)
        is set to the minimum/maximum intensities within the region.
        """

        canvasDownPos = self.getMouseDownLocation()[1]

        if canvasPos is None or canvasDownPos is None:
            return False

        overlay = self.displayCtx.getSelectedOverlay()
        if overlay is None:
            return
        display = self.displayCtx.getDisplay(overlay)
        opts    = self.displayCtx.getOpts(   overlay)
        if display.overlayType != 'volume':
            return

        copts  = canvas.opts
        x, y   = [canvasDownPos[copts.xax], canvasDownPos[copts.yax]]
        width  = canvasPos[copts.xax] - x
        height = canvasPos[copts.yax] - y

        self.__regionBriconLastRect = canvas.getAnnotations().rect(
            x, y,
            width,
            height,
            width=1,
            filled=False,
            colour=(0, 1, 1))
        canvas.Refresh()

        return True


    def _regionBriconModeLeftMouseUp(
            self, ev, canvas, mousePos, canvasPos):
        """Handles mouse up events in ``regionBricon`` mode.

        Sets the min/max display range of the currently selected overlay
        (assuming it is a ``volume``) to the minimum/maximum
        intensities within the region.
        """
        mouseDownPos, canvasDownPos = self.getMouseDownLocation()

        if canvasPos                   is None or \
           mouseDownPos                is None or \
           canvasDownPos               is None or \
           self.__regionBriconLastRect is None:
            return False

        # Clear canvas annotation
        canvas.getAnnotations().dequeue(self.__regionBriconLastRect)
        self.__regionBriconLastRect = None
        canvas.Refresh()

        # Bail if not volume
        overlay = self.displayCtx.getSelectedOverlay()
        if overlay is None:
            return
        display = self.displayCtx.getDisplay(overlay)
        opts    = self.displayCtx.getOpts(   overlay)
        if display.overlayType != 'volume':
            return

        # Turn box coordinates into voxels
        coords        = opts.transformCoords([canvasDownPos, canvasPos],
                                             'display', 'voxel', vround=True)
        coords        = coords.astype(np.int)
        xlo, ylo, zlo = coords.min(axis=0)
        xhi, yhi, zhi = coords.max(axis=0) + 1
        slc           = (slice(xlo, xhi),
                         slice(ylo, yhi),
                         slice(zlo, zhi))

        # Calculate min/max, set display range
        data = overlay.data[opts.index(slc)]

        if data.size == 0:
            return

        dmin              = np.nanmin(data)
        dmax              = np.nanmax(data)
        opts.displayRange = dmin, dmax

        return True
